structure LambdaPlusPlus =
struct
  val language_name = "Lambda Plus PLus"

  datatype exp =
    Variable of string
  | Lambda of string * exp
  | Apply of exp * exp
  | Tuple of exp * exp
  | First of exp
  | Second of exp
  | Left of exp
  | Right of exp
  | Case of exp * (string * exp) * (string * exp)
  | Fix of string * exp

  infix 6 ^^
  fun s1 ^^ s2 = s1 ^ " " ^ s2

  val rec pp_exp = (fn
      Variable x => x
    | Lambda (x, e) => "(fn" ^^ x ^^ "=>" ^^ pp_exp e ^ ")"
    | Apply (e1, e2) => "(" ^ pp_exp e1 ^^ pp_exp e2 ^ ")"
    | Tuple (e1, e2) => "(" ^ pp_exp e1 ^ "," ^^ pp_exp e2 ^ ")"
    | First e => "#1" ^^ pp_exp e
    | Second e => "#2" ^^ pp_exp e
    | Left e => "INL" ^^ pp_exp e
    | Right e => "INR" ^^ pp_exp e
    | Case (e, (x1, e1), (x2, e2)) =>
      "case" ^^ pp_exp e ^^ "of" ^^
      "INL" ^^ x1 ^^ "=>" ^^ pp_exp e1 ^^ "|" ^^
      "INR" ^^ x2 ^^ "=>" ^^ pp_exp e2
    | Fix (x, e) => "fix" ^^ x ^^ "is" ^^ pp_exp e
  )

  (* REQUIRES: e is closed *)
  fun subst (x, e) = (fn
    Variable y => if y = x then e else Variable y
  | Lambda (y, e') => Lambda (y, if y = x then e' else subst (x, e) e')
  | Apply (e1, e2) => Apply (subst (x, e) e1, subst (x, e) e2)
  | Tuple (e1, e2) => Tuple (subst (x, e) e1, subst (x, e) e2)
  | First e' => First (subst (x, e) e')
  | Second e' => Second (subst (x, e) e')
  | Left e' => Left (subst (x, e) e')
  | Right e' => Right (subst (x, e) e')
  | Case (e', (x1, e1), (x2, e2)) => Case (
      subst (x, e) e',
      (x1, if x1 = x then e1 else subst (x, e) e1),
      (x2, if x2 = x then e2 else subst (x, e) e2)
    )
  | Fix (y, e') => Fix (y, if y = x then e' else subst (x, e) e')
  )

  val evaluation_style = "eager"
  (* small-step eager dynamics and dynamic type-checking *)
  val rec step = (fn
    Variable _ => NONE
  | Lambda (x, e) => NONE
  | Apply (e1, e2) => (
      case step e1 of SOME e1' => SOME (Apply (e1', e2)) | NONE => 
      case step e2 of SOME e2' => SOME (Apply (e1, e2')) | NONE =>
      case e1 of
        Lambda (x, body) => SOME (subst (x, e2) body)
      | _ => NONE
    )
  | Tuple (e1, e2) => (
      case step e1 of SOME e1' => SOME (Tuple (e1', e2)) | NONE =>
      case step e2 of SOME e2' => SOME (Tuple (e1, e2')) | NONE => NONE
    )
  | First e => (
      case step e of SOME e' => SOME (First e') | NONE =>
      case e of
        Tuple (e1, _) => SOME e1
      | _ => NONE
    )
  | Second e => (
      case step e of SOME e' => SOME (Second e') | NONE =>
      case e of
        Tuple (_, e2) => SOME e2
      | _ => NONE
    )
  | Left e => Option.map Left (step e)
  | Right e => Option.map Right (step e)
  | Case (e, (x1, e1), (x2, e2)) => (
      case step e of SOME e' => SOME (Case (e', (x1, e1), (x2, e2))) | NONE =>
      case e of
        Left e' => SOME (subst (x1, e') e1)
      | Right e' => SOME (subst (x2, e') e2)
      | _ => NONE
    )
  | Fix (x, e) => SOME (subst (x, Fix (x, e)) e)
  )
end
